Sanity = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1", "FuBarPlugin-2.0", "AceDebug-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local tabletop = AceLibrary("TableTop-1.0")

local function Sanity_IsTBC()
	local version = GetBuildInfo();
	return string.find(version, "^2%.")
end

local function tcopy(t)
	local tt = {}
	for k,v in pairs(t) do
		tt[k] = v
	end
	setmetatable(tt, getmetatable(t))
	return tt
end

Sanity:RegisterDB("SanityDB", "SanityDBPC")
Sanity:RegisterDefaults("profile", {
	ItemHeight = 21,
	filters = {
		["SOULBOUND"] = true,
		["NOT_BOUND"] = true,
		["BAGS"] = true,
		["BANK"] = false,
		["MAIL"] = false,
		["CHAR"] = false
	},
	fieldWidths = {},
	tabletop_db = {},
	filters_shown = true,
	HighlightTexture = "Interface\Buttons\UI-Common-MouseHilight",

	UseColors = true,
	ViewBankItems = false,
	
	ToggleOnVendor = false,
	ToggleOnMail = false,
	ToggleOnBank = false,
	SearchTooltips = true,
	UseDangerousItems = false,
	bg_r = 0,
	bg_g = 0.03137,
	bg_b = 0.1333,
	Alpha = 0.86,
	UseHeaders = true,
	SortOrderBAGS = {1},
	SortOrderBANK = {1},
	SortOrderMAIL = {1},
	SortOrderCHAR = {1},
	SortOrderALL = {1},
	SaveVerbose = false,
	ShowBagIcon = true,
	UseSeparateSorts = false,
	CloseOnCombat = true,
	Scale = 1.0,
	Bagscale = 1,
	SearchOnKey = false,
	ShowCharItemTips = true,
	FocusOnShow = false,
	ItemCategories = {},
	CategoryCategories = {},
	ItemTags = {},
	CollapseStacks = true,
	ScanBag0 = true,
	ScanBag1 = true,
	ScanBag2 = true,
	ScanBag3 = true,
	ScanBag4 = true,
	ScanKeyRing = true
})


function Sanity:OnInitialize()
  -- Called when the addon is loaded
  	
  	self:Register()
	self.SortReverse = false;
	self.FilterCategories = { "BAGS", "BANK", "MAIL", "CHAR" };
	self.qualities = SANITY_QUALITIES;
	self.activeCategory = {"BAGS"}
	self.sortTypes = {SANITY_COLUMNS_LEVEL = "number"}
	
	self.SecondarySortReverse = {
		["Quality"] = 1,
		["MinLevel"] = 1
	}

	self.splitBag = -1;
	self.splitIndex = -1;
	self.lastSort = {"Name"};
	self.filterString = "";
	self.activeItemList = {}
	self.itemList = {}
	self.character = UnitName("player")
	self.defaultColor = {242/255, 198/255, 3/255};
	self.needMailboxScan = false;
	self.ItemCache = SanityItemCache;
	self.filter_soulbound = 0;
	self.viewList = {}
	self.visibleItems = 0
	
	self.BagSlots = 0
	self.BankSlots = 0
	self.UsedBagSlots = 0
	self.UsedBankSlots = 0
	self.tabName = "SanityTabButton"
	self.tabletop = tabletop
	self.TooltipDataCache = {}
	self.displayListDirty = true
	
	Sanity.DraggerWidth = 10
	
	-- From MyBags
	self.SLOTCOLOR     = { 0.5, 0.5, 0.5 }
	self.AMMOCOLOR     = { 0.6, 0.6, 0.1 }
	self.SHARDCOLOR    = { 0.6, 0.3, 0.6 }
	self.ENCHANTCOLOR  = { 0.2, 0.2, 1.0 }
	self.ENGINEERCOLOR = { 0.6, 0.0, 0.0 }
	self.HERBCOLOR     = { 0.0, 0.6, 0.0 }
	self.KEYRINGCOLOR  = { 0.7, 0.4, 0.4 }
	
	self.BankBagsIndexes = { -2, -1, 5, 6, 7, 8, 9, 10, 0, 1, 2, 3, 4 };
	self.BankIndexes = { 10, 9, 8, 7, 6, 5, -1 };
	self.BagsIndexes = { -2, 4, 3, 2, 1, 0 };
	
	self.bagFilters = {
		{SANITY_TABS_BAG, "BAGS", function() end}, 
		{SANITY_TABS_BANK, "BANK"},
		{SANITY_TABS_MAIL, "MAIL"},
		{SANITY_TABS_CHAR, "CHAR"}
	}
	self.bindFilters = {
		{SANITY_FILTER_SOULBOUND, "SOULBOUND"},
		{SANITY_FILTER_NOTBOUND, "NOT_BOUND"}
	}
	
	self.owner_keys = {
		{"BagItems", SANITY_TABS_BAG}, 
		{"BankItems", SANITY_TABS_BANK}, 
		{"EquipmentItems", SANITY_TABS_CHAR}, 
		{"MailItems", SANITY_TABS_MAIL}
	}
	
	self.BUTTONHEIGHT = 16;
	self.ItemColors = {
		{157/255, 157/255, 157/255},
		{255/255, 255/255, 255/255},
		{30/255, 255/255, 0/255},
		{0/255, 112/255, 221/255},
		{163/255, 53/255, 238/255},
		{255/255, 128/255, 0/255}
	};
	
	self.ItemLinkColors = {
		"dddddddd",		-- Trash
		"ffffffff",		-- Common
		"ff0070dd",		-- Rare
		"ff1eff00",		-- Uncommon
		"ffa335ee",		-- Epic
	}
	
	SanityFrame:SetMinResize(295, 47)

	Sanity:SetupFilters(self.bagFilters, SanityTabsFrame)
	Sanity:SetupFilters(self.bindFilters, SanityBindFrame)

	tabletop:Init(Sanity.db.profile.tabletop_db, SanitySortFrame, SanityFrame)
	tabletop:SetDataRowOffsets({
		tab_left = 5,
		tab_top = 0,
		tab_right = -4,
		tab_bottom = 0,
		data_left = 5,
		data_top = -5,
		data_right = -4,
		data_bottom = 0
		})
	tabletop:LayoutTabs(self.columns, 
		"SanityHeaderTabTemplate", 
		function()
			Sanity:PrintLockWarning()
			if not InCombatLockdown() then
				Sanity.displayListDirty = true
				if this.tabdata.key == "MinLevel" then
					Sanity:Sort({this.tabdata.key, "Quality", "Name"}, true)
				else
					Sanity:Sort({this.tabdata.key, "Name"}, true)
				end
				self.tabletop:ResizeTabContainer()
			end
		end 
	)
	Sanity:SetupItems()
	tabletop:LayoutInit()
	if not Sanity.db.profile.ShowBagIcon then
		SanityBagButton:Hide()
	end
end

function Sanity:OnEnable()
	self:RegisterEvent("BAG_UPDATE");
	self:RegisterEvent("BANKFRAME_CLOSED");
	self:RegisterEvent("BANKFRAME_OPENED");
	self:RegisterEvent("MERCHANT_SHOW");
	self:RegisterEvent("MERCHANT_CLOSED");
	self:RegisterEvent("MAIL_SHOW");
	self:RegisterEvent("MAIL_INBOX_UPDATE")
	self:RegisterEvent("MAIL_CLOSED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	local funcs = {"SetAuctionItem","SetAuctionSellItem","SetBagItem","SetBuybackItem","SetCraftItem","SetHyperlink",
		"SetInboxItem","SetInventoryItem","SetLootItem","SetLootRollItem","SetMerchantCompareItem","SetMerchantItem",
		"SetQuestItem","SetQuestLogItem","SetSendMailItem","SetTradePlayerItem","SetTradeTargetItem"}
	for k,v in pairs(funcs) do
		if not self:IsHooked(GameTooltip, v) then
			self:SecureHook(GameTooltip, v, Sanity.AddTooltipInfo);
		end
	end
	
	SanityItemCache:RegisterEvent("BAG_UPDATE")
	SanityItemCache:RegisterEvent("PLAYER_MONEY")
	SanityItemCache:RegisterEvent("BANKFRAME_OPENED")
	SanityItemCache:RegisterEvent("UNIT_INVENTORY_CHANGED")	
	SanityItemCache:BAG_UPDATE()
	self.IsEnabled = true
end

function Sanity:OnDisable()
	SanityBagButton:Hide()
	SanityItemCache:UnregisterEvent("BAG_UPDATE")
	SanityItemCache:UnregisterEvent("PLAYER_MONEY")
	SanityItemCache:UnregisterEvent("BANKFRAME_OPENED")
	SanityItemCache:UnregisterEvent("UNIT_INVENTORY_CHANGED")	
	self.IsEnabled = false
end

function Sanity:AddTooltipInfo()
	if not Sanity.db.profile.ShowCharItemTips then return end
	local name = GameTooltipTextLeft1:GetText()
	if not name then return end

	if GetMouseFocus() ~= Sanity.lastFocus then
		Sanity.lastFocus = nil
		Sanity.lastFocusName = nil
		local b = getglobal(GetMouseFocus():GetName() .. "IconTexture")
		if b then
			Sanity.lastFocus = GetMouseFocus()
			Sanity.lastFocusTexture = b:GetTexture()
			Sanity.lastFocusName = name
		end
	end

	local owners = Sanity:GetOwnersFor(name)
	if not owners then return end
	local tct = 0
	local ctct = 0
	for char,v in pairs(owners) do
		local str = char .. " has "
		local i = 0
		for loc, ct in pairs(v) do
			if i ~= 0 then
				str = str .. ", "
			end
			i = i + 1
			tct = tct + ct
			ctct = ctct + 1
			str = str .. "[" .. loc .. ": " .. ct .. "]"
		end
		GameTooltip:AddLine(str, 0.3, 0.75, 1, 1)
	end
	if ctct > 1 then
		GameTooltip:AddLine("Total: " .. tct, 0.1, 0.75, 1, 1)
	end
	str = Sanity:GetCategoryFor(name)
	if str then
		GameTooltip:AddLine("Category: " .. str, 0.1, 0.85, 0.6, 1)
	end
	str = Sanity:GetTagsFor(name)
	if str ~= "" then
		GameTooltip:AddLine("Tags: " .. str, 0.1, 0.85, 0.2, 1)
	end
	if Sanity.lastFocusName and Sanity.lastFocus and GetBindingKey("SANITYITEMMENU") then
		GameTooltip:AddLine("<" .. GetBindingKey("SANITYITEMMENU") .. "> to tag this item.")
	end
	GameTooltip:Show()
end

function Sanity:DropItemOnCategory(category)
	local header = getglobal(category:GetName() .. "SECTION_HEADER")
	local lockedItem = nil
	local catName = header:GetText()
	if header:IsVisible() then
		for k, v in pairs(self.BankBagsIndexes) do
			for slot = 1, GetContainerNumSlots(v) do
				local texture, itemCount, locked, quality, readable = GetContainerItemInfo(v, slot);
				if locked and not lockedItem then
					lockedItem = {v, slot}
				elseif locked and lockedItem then
					self:Print("Unable to categorize cursor item: Multiple items are locked.")
					return
				end
			end
		end
	end
	if lockedItem then
		local link = GetContainerItemLink(lockedItem[1], lockedItem[2])
		if not item.invalid then
			local itemName, itemString, itemQuality, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(link)
			if not itemName then itemName = "Unknown" end
			self:SetItemCustomCategory(itemName, catName)
		else
			self:SetItemCustomCategory("Unknown", catName)
		end
		ClearCursor()
	end
end
                             
function Sanity:SetItemCustomCategory(item, cat)
	Sanity.db.profile.ItemCategories[item] = cat
	Sanity:Refresh()
end

function Sanity:GetTagList()
	if Sanity.tagList and not Sanity.tagsDirty then return Sanity.tagList end
	local tags = {}
	for k,v in pairs(Sanity.db.profile.ItemTags) do
		for k1,v1 in pairs(v) do
			tags[k1] = 1
		end
	end
	local tList = {}
	for k,v in pairs(tags) do
		tinsert(tList, tostring(k))
	end
	table.sort(tList)
	Sanity.tagList = tList
	Sanity.tagsDirty = false
	return tList
end

function Sanity:GetLinkID(link)
	if Sanity_IsTBC() then return link end
	local itemId = nil;
	if ( type(link) == "string" ) then
		_,_, itemId = string.find(link, "(item:%d+:%d+:%d+:%d+)");
		return itemId;
	end
end

function Sanity:GetTagsFor(item)
	if not Sanity.db.profile.ItemTags[item] then return "" end
	local s = ""
	local tags = {}
	for k,v in pairs(Sanity.db.profile.ItemTags[item]) do
		tinsert(tags, k)
	end
	table.sort(tags)
	for k,v in pairs(tags) do
		if s == "" then
			s = v
		else
			s = s .. ", " .. v
		end
	end
	return s
end

function Sanity:GetCategoryFor(item)
	local name = self.db.profile.ItemCategories[item]
	if name then
		local catName = self.db.profile.CategoryCategories[name]
		if catName then return catName end
	end
	return name
end

function Sanity:ItemHasTag(item, tag)
	if Sanity.db.profile.ItemTags[item] then
		return Sanity.db.profile.ItemTags[item][tag] ~= nil
	end
	return false
end

function Sanity:TagItem(item, tag)
	if not Sanity.db.profile.ItemTags[item] then
		Sanity.db.profile.ItemTags[item] = {}
	end
	Sanity.db.profile.ItemTags[item][tag] = 1
	Sanity.tagsDirty = true
	Sanity:GetTagList()
	Sanity:Refresh()
end

function Sanity:UntagItem(item, tag, noRefresh)
	if not Sanity.db.profile.ItemTags[item] then return end
	local tTags = {}
	for k,v in pairs(Sanity.db.profile.ItemTags[item]) do
		if k ~= tag then
			tTags[k] = 1
		end
	end
	Sanity.db.profile.ItemTags[item] = tTags
	if not noRefresh then
		Sanity.tagsDirty = true
		Sanity:GetTagList()
		Sanity:Refresh()
	end
end

function Sanity:GetItemsForTag(tag)
	if not Sanity.db.profile.ItemTags then return end
	local result = {}
	local resultCount = 0
	for k,v in pairs(Sanity.db.profile.ItemTags) do
		for k2, v2 in pairs(v) do
			if k2 == tag then
				tinsert(result, k)
			end
		end
	end
	return result
end

function Sanity:KillTag(tag)
	if not Sanity.db.profile.ItemTags then return end
	for k,v in pairs(Sanity.db.profile.ItemTags) do
		Sanity:UntagItem(k, tag, true)
	end
	Sanity.tagsDirty = true
	Sanity:GetTagList()
	Sanity:Refresh()
end
 
function Sanity:GetOwnersFor(itemName)
	if itemName == self.lastScanName and not self.tooltipDirty then
		return self.lastScanResults
	end
	self.tooltipDirty = false
	local results = {}
	for charName, char in pairs(SanityItemCache.db.realm.chars) do
		for k, key in pairs(self.owner_keys) do
			if char[key[1]] then
				for i, item in pairs(char[key[1]]) do
					if item["Name"] == itemName then
						if not results[charName] then results[charName] = {} end
						if not results[charName][key[2]] then results[charName][key[2]] = 0 end
						results[charName][key[2]] = results[charName][key[2]] + item["Count"]
					end
				end
			end
		end		
	end
	self.lastScanResults = results
	self.lastScanName = itemName
	return results
end

function Sanity:UpdateColumns(tabs)
	if type(tabs) ~= "table" then return end 
	for k,tab in pairs(tabs) do
		Sanity.db.profile.fieldWidths[tab[1]] = tab[2]
	end
end

function Sanity:SetupFilters(tabs, parent)
	local name = parent:GetName() .. "FilterCheck"
	for k,v in pairs(tabs) do
		local f = CreateFrame("Button", name .. k, parent, "SanityFilterCheckboxTemplate")
		f:SetText(v[1])
		f.value = v[2]
		f.group_name = name
		f.group_size = table.getn(tabs)
		if k == 1 then
			f:SetPoint("TOPLEFT", parent, "TOPLEFT", 7, -5)	
		else
			f:SetPoint("TOPLEFT", getglobal(name .. (k-1)), "BOTTOMLEFT", 0, 0)
		end
		f:SetScript("OnClick", function()
				Sanity.displayListDirty = true;
				for i=1,this.group_size do
					local checkbox = getglobal(this.group_name .. i .. "Checkbox")
					checkbox:SetChecked(false)
				end
				getglobal(this:GetName() .. "Checkbox"):SetChecked(true)
				Sanity:Refresh()
			end)
		local checkbox = getglobal(name .. k .. "Checkbox")
		checkbox:SetChecked(Sanity.db.profile.filters[v[2]])
		checkbox:SetScript("OnClick", function() Sanity.displayListDirty = true; Sanity:Refresh() end)
	end
	parent:SetHeight(table.getn(tabs) * 18 + 8)
end

function Sanity:GCDebug(label, reset)
	if reset then Sanity.lastGC = nil end
	if Sanity.lastGC == nil then
		Sanity.lastGC = gcinfo()
	else
		local g1 = gcinfo()
		Sanity:Debug(label .. ": " .. (g1 - Sanity.lastGC))
		Sanity.lastGC = nil
	end
end

function Sanity:GetTooltipString(link)
	if not link then return nil end
	if not Sanity.db.profile.SearchTooltips then return nil end
	local name = GetItemInfo(link)
	if not name then name = "Unknown" end
	if self.TooltipDataCache[name] then
		return self.TooltipDataCache[name]
	end
	Sanity_ItemTooltip:ClearLines()
	Sanity_ItemTooltip:SetOwner(WorldFrame, "ANCHOR_NONE")
	Sanity_ItemTooltip:SetHyperlink(Sanity:GetLinkID(link))
	local s = nil
	for i=1,Sanity_ItemTooltip:NumLines() do
		local lText = getglobal("Sanity_ItemTooltipTextLeft" .. i)
		local rText = getglobal("Sanity_ItemTooltipTextRIght" .. i)
		if lText and lText:GetText() then
			if s then s = s .. " " .. lText:GetText() 
			else s = lText:GetText() end
		end
		if rText and rText:GetText() then
			if s then s = s .. " " .. rText:GetText() 
			else s = rText:GetText() end
		end
	end
	self.TooltipDataCache[name] = s
	return s
end

function Sanity:SetupItems()
	local ht = SanityFrame:GetHeight() - 9
	Sanity.tabletop:GetColumn("TypeSort"):SetWidth(Sanity.db.profile.ItemHeight)
	Sanity.visibleItems = math.floor(ht / Sanity.db.profile.ItemHeight)
	Sanity.tabletop:SetupRows({
		parent = SanityFrame,
		num = Sanity.visibleItems,
		height = Sanity.db.profile.ItemHeight,
		template = "SanityItemTemplate"
	})
	Sanity:Refresh(true)
end

function Sanity:SetItemCategory(val)
	if dewdrop:GetOpenedParent() then
		local item = Sanity.viewList[tonumber(getglobal(dewdrop:GetOpenedParent():GetParent():GetName() .. "ItemIndex"):GetText())]
		if item["Header"] then
			local haveSet = false
			for k,v in pairs(Sanity.db.profile.CategoryCategories) do
				if v == item["Header"] then
					Sanity.db.profile.CategoryCategories[k] = val
					haveSet = true
				end
			end
			if not haveSet then
				Sanity.db.profile.CategoryCategories[item["Header"]] = val
			end
			for k,v in pairs(Sanity.db.profile.ItemCategories) do
				if k == item["Header"] then
					Sanity.db.profile.ItemCategories[k] = val
				end
			end
		else
			Sanity.db.profile.ItemCategories[item["Name"]] = val
		end
	end			
	Sanity:Refresh()
	dewdrop:Close()
end

function Sanity:GetExistingCategories()
	local existingCategoryNames = {}
	local existingCategories = {}
	local i = 0
	for k,v in pairs(Sanity.db.profile.ItemCategories) do
		if v then
			existingCategoryNames[v] = 1
		end
	end
	for k,v in pairs(Sanity.db.profile.CategoryCategories) do
		if v then
			existingCategoryNames[v] = 1
		end
	end
	local cats = {}
	for k,v in pairs(existingCategoryNames) do
		tinsert(cats, k)
	end
	return cats
end

function Sanity:UpdateBackground()
	local r = Sanity.db.profile.bg_r
	local g = Sanity.db.profile.bg_g
	local b = Sanity.db.profile.bg_b
	local a = Sanity.db.profile.Alpha
	local frames = {SanityFrame, SanityTabsFrame, SanityBindFrame, SanitySortFrame,
									SanityHeaderFrame, SanityTitleFrame, SanityStatusBar, SanityFiltersFrame}
	for idx, frame in pairs(frames) do
		frame:SetBackdropColor(r,g,b,a);
	end
end

function Sanity:Toggle()
	if SanityFrame:IsVisible() then
		SanityFrame:Hide();
	else
		-- SanityItemCache.enable_collection = true
		Sanity.displayListDirty = true
		SanityItemCache:BAG_UPDATE()
		SanityItemCache:PLAYER_MONEY()
		
		SanityFrame:Show()
		if (not Sanity.db.profile.filters_shown) and SanityTabsFrame:IsVisible() then
			Sanity:ToggleFilters(true)
		end
		MoneyFrame_Update("SanityStatusBarMoneyFrame", SanityItemCache.db.realm.chars[Sanity.character].Money)
		Sanity:UpdateBackground()
		Sanity:SetupItems()
		Sanity.tabletop:ResizeTabContainer()
		self:Refresh()
		if self.db.profile.FocusOnShow then
			SanityEditBox:HighlightText();
			SanityEditBox:SetFocus();
		end
	end
end

function Sanity:PrintLockWarning()
	if InCombatLockdown() and not Sanity.printedWarning then
		Sanity:Print("Not modifiable while in combat. Try again when OOC.")
		Sanity.printedWarning = true
	end
end

function Sanity:EditBoxUpdate()
	Sanity:PrintLockWarning()

	if not InCombatLockdown() and this:GetText() ~= self.filterString then
		self.displayListDirty = true
		self.filterString = this:GetText();
		self:Refresh();
	end
end

function Sanity:Compress()
	self.DoCompress = true;
end

function Sanity:GetFreeBag()
	for bag_id in self.BagsIndexes do
		if GetContainerNumSlots(bag_id) then
			for slot_id = 1, GetContainerNumSlots(bag_id) do
				if not GetContainerItemLink(bag_id, slot_id) then
					return bag_id, slot_id
				end
			end
		end	
	end
end

function Sanity:GetBags()
	return SanityItemCache.db.realm.chars[self.character].BagItems
end

function Sanity:GetBank()
	return SanityItemCache.db.realm.chars[self.character].BankItems
end

function Sanity:GetEquipment()
	return SanityItemCache.db.realm.chars[self.character].EquipmentItems
end

function Sanity:GetMailbox()
	return SanityItemCache.db.realm.chars[self.character].MailItems
end

function Sanity:Refresh(skip_list_rebuild)
	if not SanityFrame:IsVisible() then return end
	if not skip_list_rebuild or not self.itemList then
		local baglist = {}
		local lists = {}
		Sanity.activeCategory = {}
		local checkCount = 0
		self.filter_soulbound = 0
		for i = 1,table.getn(Sanity.bindFilters) do
			local obj = getglobal(SanityBindFrame:GetName() .. "FilterCheck" .. i)
			local check = getglobal(obj:GetName() .. "Checkbox")
			Sanity.db.profile.filters[obj.value] = false
			if check:GetChecked() then
				Sanity.db.profile.filters[obj.value] = true
				checkCount = checkCount + 1
				if obj.value == "SOULBOUND" then
					self.filter_soulbound = 1
				elseif obj.value == "NOT_BOUND" then
					self.filter_soulbound = 2
				end
			end
		end
		if checkCount == 2 then
			self.filter_soulbound = 0
		end
		
		for i = 1,table.getn(Sanity.bagFilters) do
			local obj = getglobal(SanityTabsFrame:GetName() .. "FilterCheck" .. i)
			local check = getglobal(obj:GetName() .. "Checkbox")
			Sanity.db.profile.filters[obj.value] = false
			if check:GetChecked() then
				Sanity.db.profile.filters[obj.value] = true
				tinsert(Sanity.activeCategory, obj.value)
			end
		end
		for k,v in pairs(Sanity.FilterCategories) do
			for k2,v2 in pairs(Sanity.activeCategory) do
				if v == v2 then
					if v == "BAGS" then
						tinsert(lists, Sanity:GetBags())
					elseif v == "BANK" then
						tinsert(lists, Sanity:GetBank())
					elseif v == "CHAR" then
						tinsert(lists, Sanity:GetEquipment())
					elseif v == "MAIL" then
						tinsert(lists, Sanity:GetMailbox())
					end
				end
			end
		end
		self.itemList = {}
		for index, list in pairs(lists) do
			for index, item in pairs(list) do
				if item then
					item["Owner"] = self.character
					tinsert(self.itemList, item)
				end
			end
		end
	end
	local list = self:GetDisplayList()
	self:UpdateDisplay(list)
end

function Sanity:ResetSanity()
	StaticPopup_Show("SANITY_CONFIRM_DATABASE_WIPE")
end

function Sanity:GetDisplayList()
	if Sanity.lastDisplayList and table.getn(Sanity.lastDisplayList) > 0 and not Sanity.displayListDirty then
		return Sanity.lastDisplayList
	end
	Sanity:Debug("List is dirty, getting new list")
	self.TooltipDataCache = {}
	
	self:PrepareList()

	Sanity.displayListDirty = false
	if (self.filterString and self.filterString ~= "") or self.filter_soulbound > 0 then
		local tempList = {}
		for idx, item in pairs(self.itemList) do
			local cSort = item[Sanity.lastSort[1]]
			if Sanity.lastSort[1] == "Quality" then
				cSort = SANITY_QUALITIES[cSort+1]
			end
			
			if string.find(string.upper(item["Name"]), string.upper(self.filterString), 0, true) ~= nil or
				string.find(string.upper(cSort), string.upper(self.filterString), 0, true) ~= nil or 
				(item["Tags"] and string.find(string.upper(item["Tags"]), string.upper(self.filterString), 0, true) ~= nil) or 
				(Sanity.db.profile.SearchTooltips and string.find(string.upper(item["Tooltip"]), string.upper(self.filterString), 0, true) ~= nil) or 
				(item["TypeSort"] and string.find(string.upper(item["TypeSort"]), string.upper(self.filterString), 0, true) ~= nil) then
				if self.filter_soulbound == 0 or (self.filter_soulbound == 1 and item["Soulbound"] == 1) or (self.filter_soulbound == 2 and item["Soulbound"] == 0) then
					tinsert(tempList, item)
				end
			end			
		end
		self.itemList = tempList
	end
	
	local headers = {}
	if Sanity.db.profile.UseHeaders then
		headers = self:GetHeaders()
	end

	local displayList = {}
	for k,v in pairs(self.itemList) do
		tinsert(displayList, v)
	end
	for k,v in pairs(headers) do
		tinsert(displayList, v)
	end
	
	local result = self:SortList(displayList)

	Sanity.lastDisplayList = result
	return result
end

function Sanity:GetHeaders()
	local lastSortHeading = "";
	local sort = self.lastSort[1]
	local headers = {}
	setmetatable(headers, {__mode = "v"})
	if sort == "Name" then return headers end
	
	if sort == "TypeSort" then
		local categories = {}
		for k,v in pairs(self.itemList) do
			categories[tostring(v["TypeSort"])] = 1
		end
		for k,v in pairs(categories) do
			tinsert(headers, {
				["Name"] = "AAAAAAAAA",
				["Type"] = k,
				["Header"] = k,
				["TypeSort"] = k,
				["Priority"] = 1
			})
		end
		return headers
	elseif sort == "Quality" then
		local categories = {}
		for k,v in pairs(self.itemList) do
			categories[self.qualities[v["Quality"]+1]] = v["Quality"]
		end
		for k,v in pairs(categories) do
			tinsert(headers, {
				["Name"] = "AAAAAAAAA",
				["Quality"] = v,
				["Header"] = k,
				["Priority"] = 1
			})
		end
		return headers		
	else
		local categories = {}
		for k,v in pairs(self.itemList) do
			categories[tostring(v[sort])] = 1
		end
		for k,v in pairs(categories) do
			local name = "AAAAAAAAAAAAA"
			if self.sortTypes[k] == "number" then
				name = "000000000000000000000"
			end
			tinsert(headers, {
				["Name"] = name,
				[sort] = k,
				["Header"] = k,
				["Priority"] = 1
			})
		end
		return headers		
	end
	
	return headers
end

function Sanity:PrepareList()
	Sanity:Debug("Preparing item list")
	local tempListIndexes = {}
	local tempList = {}
	for idx, item in pairs(self.itemList) do
		local skip = false
		if not item["Location"] and item["Link"] then
			item["Location"] = SANITY_TABS_BANK
			if item["Bag"] == SanityItemCache.INVENTORY_BAG_ID then
				item["Location"] = SANITY_TABS_CHAR
			end
			for k,v in pairs(self.BagsIndexes) do
				if v == item["Bag"] then
					item["Location"] = SANITY_TABS_BAG
				end
			end
		elseif not item["Location"] then
			item["Location"] = SANITY_TABS_MAIL
		end
		local itemKey = item["Name"] .. "|" .. item["Location"]
		if not item["Prepped"] then
			item["Prepped"] = true
			if Sanity.db.profile.CollapseStacks then
				if tempListIndexes[itemKey] then
					Sanity:Debug("Merging " .. item["Name"] .. " with existing stack")
					self.itemList[tempListIndexes[itemKey]]["Count"] = self.itemList[tempListIndexes[itemKey]]["Count"] + item["Count"]
					self.itemList[tempListIndexes[itemKey]]["Stacks"] = self.itemList[tempListIndexes[itemKey]]["Stacks"] + 1
					item["Skip"] = true
					skip = true
				end
			end
			if not skip then
				tempListIndexes[itemKey] = idx
				item["Stacks" ] = 1
				if item["Link"] then
					if item["Dangerous"] then
						item["Dangerous"] = false
					end	
					if not item["Type"] then
						item["Type"] = SANITY_UNKNOWN
						item["SubType"] = SANITY_UNKNOWN
						item["EquipLoc"] = SANITY_UNKNOWN
						item["Dangerous"] = true
					end
				else
					item["Type"] = "Mailbox"
					item["SubType"] = "Mailbox"
					item["EquipLoc"] = "nil"
					item["Soulbound"] = false
				end	
				item["Tags"] = Sanity:GetTagsFor(item["Name"])
				if not item.invalid then
					item["Tooltip"] = Sanity:GetTooltipString(item["Link"]) or ""
				else
					item["Tooltip"] = ""
				end
				item["TypeSort"] = "zzz"
				local catName = self.db.profile.ItemCategories[item["Name"]]
				if catName then
					item["TypeSort"] = catName
				else
					if item["Type"] == item["SubType"] then
						if item["EquipLoc"] ~= "nil" then
							item["TypeSort"] = item["SubType"] .. " - " .. item["EquipLoc"]
						else
							item["TypeSort"] = item["SubType"]
						end
					elseif item["EquipLoc"] ~= "nil" then
						item["TypeSort"] = item["Type"] .. " (" .. item["SubType"] .. ") - " .. item["EquipLoc"]
					else
						item["TypeSort"] = item["Type"]
					end
				end
				local catCatName = self.db.profile.CategoryCategories[item["TypeSort"]]
				if catCatName then
					item["TypeSort"] = catCatName
				end
				if item["TypeSort"] == nil then
					item["TypeSort"] = "No Type"
				end
				
				if item["Bag"] and not getglobal("SanityDummyBagFrame" .. item["Bag"]) then
					local f = CreateFrame("Frame", "SanityDummyBagFrame" .. item["Bag"], SanityFrame)
					f:SetID(item["Bag"])
				end
				tinsert(tempList, item)
				-- self.itemList[idx] = item
			end
		elseif not Sanity.db.profile.CollapseStacks or Sanity.db.profile.CollapseStacks and not tempListIndexes[itemKey] then
			tinsert(tempList, item)
			tempListIndexes[itemKey] = idx
		end
	end
	self.itemList = tempList
end

-- This is called from Refresh, so we don't need to call it directly. Just modify the sort indexes and call Refresh.
function Sanity:SortList(list)
	if (SanityFrame:IsVisible() == false) or self.lastSort == nil then
		return list;
	end
	local indexT = self.lastSort;
	table.sort(list, 
		function (t1, t2)
			local total = 0;
			local r1 = 0;
			local r2 = 0;
			for x,j in pairs(indexT) do
				local i = j;
				local s = math.pow(2, table.getn(indexT)-x);
				if not (t1 == nil or t2 == nil or t1[i] == nil or t2[i] == nil) then
					local i1 = t1[j]
					local i2 = t2[j]
					if tonumber(i1) and tonumber(i2) then
						i1 = tonumber(i1)
						i2 = tonumber(i2)
					end
					if (self.SecondarySortReverse[j] and x ~= 1) or
						 (x == 1 and self.SortReverse and not self.SecondarySortReverse[j]) or 
						 (x == 1 and not self.SortReverse and self.SecondarySortReverse[j]) then
						if i1 > i2 then
							r1 = r1 + s
						elseif i1 < i2 then
							r2 = r2 + s;
						end
					else
						if i1 < i2 then
							r1 = r1 + s
						elseif i1 > i2 then
							r2 = r2 + s;
						end
					end
				end
			end
			return (r1 > r2);
		end
	);	
	return list
end

function Sanity:Sort(indexT, reverse)
	if indexT == nil then indexT = lastSort; end
	if type(indexT) ~= "table" then
		indexT = {}
	end
	
	if indexT[1] == "Name" and table.getn(indexT) > 1 then
		table.remove(indexT, 1);
	end
	
	if self.lastSort[1] == indexT[1] then
		self.SortReverse = not self.SortReverse
	else
		self.SortReverse = false
	end
	
	self.lastSort = indexT;
	Sanity:Refresh();
end

function Sanity:GetHeaderForSortType(sortType, list, index)
	if sortType == "Quality" then
		return quality[list[index][sortType]];
	end
	return list[index][sortType]
end
