local dewdrop = AceLibrary("Dewdrop-2.0")
local tablet = AceLibrary("Tablet-2.0")
local crayon = AceLibrary("Crayon-2.0")
local debug = AceLibrary("AceDebug-2.0")

Sanity.columns = {
	{
		name 		= SANITY_COLUMNS_TYPE,
		key 		= "TypeSort",
		width 		= 35,
		fixed		= true,
		type		= "button",
		inherits	= "SanityItemIcon",
		desc 		= "Item's category. This includes any custom categories you set."
	},
	{
		name 		= SANITY_COLUMNS_NAME,
		key			= "Name",
		desc 		= "Item name."
	},
	{
		name 		= SANITY_COLUMNS_QUALITY,
		key 		= "Quality",
		nocolumn 	= true,
		desc		= "Item's quality"
	},
	{
		name 		= SANITY_COLUMNS_LEVEL,
		nocolumn 	= true,
		minwidth 	= 18,
		key 		= "MinLevel",
		desc		= "Minimum level required to equip or use this item."
	},
	{
		name 		= SANITY_COLUMNS_STACKS,
		minwidth 	= 35,
		key 		= "Stacks",
		desc		= "Number of stacks of this item you possess."
	},
	{
		name 		= SANITY_COLUMNS_TAGS,
		key 		= "Tags",
		minwidth 	= 18,
		hidden		= true,
		desc			= "Tags you've assigned to this item."
	},
	{
		name 		= SANITY_COLUMNS_OWNER,
		key 		= "Owner",
		desc		= "Character that owns this item."
	},
	{
		name 		= SANITY_COLUMNS_LOCATION,
		key 		= "Location",
		minwidth 	= 75,
		desc		= "Where this item is located (bags, bank, etc)."
	}
}

local options = {
    type='group',
    args = {
        show = {
        	type = "execute",
        	desc = "Toggle the main Sanity window",
        	name = "Toggle Sanity",
        	func = "Toggle",
        	order = 102
        },

        toggleat = {
        	type = "group",
        	name = "Toggle At...",
        	desc = "Cause Sanity to automatically toggle on and off on certain events.",
        	args = {
		        vendortoggle = {
		            type = 'toggle',
		            name = 'Toggle At Vendors',
		            desc = 'Open Sanity when you speak to a vendor',
		            get = function()
		            	return Sanity.db.profile.ToggleOnVendor
		            end,
		            set = function(v)
		            	Sanity.db.profile.ToggleOnVendor = v
		            end
		        },
		        banktoggle = {
		            type = 'toggle',
		            name = 'Toggle At Bank',
		            desc = 'Open Sanity when you speak to a banker',
		            get = function()
		            	return Sanity.db.profile.ToggleOnBank
		            end,
		            set = function(v)
		            	Sanity.db.profile.ToggleOnBank = v
		            end
		        },
		    		mailtoggle = {
		            type = 'toggle',
		            name = 'Toggle At Mailboxes',
		            desc = 'Open Sanity when you open a mailbox',
		            get = function()
		            	return Sanity.db.profile.ToggleOnMail
		            end,
		            set = function(v)
		            	Sanity.db.profile.ToggleOnMail = v
		            end
		        }, 		        
		      }
	      },
	      bagbutton = {
        	type = "group",
        	name = "Bag Button Options",
        	desc = "Options relating to the bag button.",
        	args = {
		        showbag = {
		            type = 'toggle',
		            name = 'Show Bag Button',
		            desc = "Show the Sanity bag button.",
		            get = function()
		            	return Sanity.db.profile.ShowBagIcon
		            end,
		            set = function(v)
		            	Sanity.db.profile.ShowBagIcon = v
		            	if v then
		            		SanityBagButton:Show()
		            	else
		            		SanityBagButton:Hide()
		            	end
		            end
		        },        
		        bagscale = {
		            type = 'range',
		            name = 'Bag Icon Scale',
		            step = 0.05,
		            min = 0.4,
		            max = 5.0,
		            isPercent = true,
		            desc = "Changes the scale of the Sanity bag icon",
		            get = function()
		            	return Sanity.db.profile.Bagscale
		            end,
		            set = function(v)
		            	if v ~= Sanity.db.profile.Bagscale then
										SanityBagButton:SetScale(v);
										SanityBagButton:ClearAllPoints();
										SanityBagButton:SetPoint("CENTER", "UIParent", "CENTER");
			            	Sanity.db.profile.Bagscale = v
			            end
		            end
		        }                	
        	}
	      },
	      tooltipoptions = {
        	type = "group",
        	name = "Tooltip Options",
        	desc = "Options relating to tooltips.",
        	args = {
		        searchtips = {
		            type = 'toggle',
		            name = 'Search Tooltips',
		            desc = "Includes tooltip data in searches. This lets you perform highly advanced searches, but may slow you down. Turn this off if Sanity is slow, or you don't want the advanced searches.",
		            get = function()
		            	return Sanity.db.profile.SearchTooltips
		            end,
		            set = function(v)
		            	Sanity.db.profile.SearchTooltips = v
		            end
		        },            	
		        show_tooltip_info = {
		        	type = 'toggle',
		        	desc = "Show which of your characters has how many of an item when you mouse over it",
		        	name = "Show Item Tooltip Information",
		          get = function()
		          	return Sanity.db.profile.ShowCharItemTips
		          end,
		          set = function(v)
		          	Sanity.db.profile.ShowCharItemTips = v
		          end
		        },   		        
        	}
	      },
        advancedoptions = {
        	type = "group",
        	name = "Advanced Options",
        	desc = "Advanced Options",
        	order = 101,
        	args = {
		        deletesettings = {
		        		type = "execute",
		        		func = "ResetSanity",
		        		name = "Reset Database",
		        		desc = "Resets all of Sanity's settings and reloads your UI."
		        },
		        usedangerous = {
		            type = 'toggle',
		            name = 'Use Dangerous Item Code',
		            desc = "Attempts to pull data on unknown items from the server. However, if the server hasn't seen this item since its last restart, you will be disconnected. This is off by default.",
		            get = function()
		            	return Sanity.db.profile.UseDangerousItems
		            end,
		            set = function(v)
		            	Sanity.db.profile.UseDangerousItems = v
		            end
		        },
        	}
        },
        displayoptions = {
        	type = "group",
        	name = "Display Options",
        	desc = "Modify Sanity's look and feel",
        	args = {
		        collapsestacks = {
		            type = 'toggle',
		            name = 'Collapse Multiple Stacks',
		            desc = 'Collapses multiple stacks into a single stack for visual clarity.',
		            get = function()
		            	return Sanity.db.profile.CollapseStacks
		            end,
		            set = function(v)
		            	Sanity.db.profile.CollapseStacks = v
		            	SanityItemCache:BAG_UPDATE()
		            	Sanity.displayListDirty = true
		            	Sanity:Refresh()
		            end
		        },       
		        closeoncombat = {
		            type = 'toggle',
		            name = 'Close On Combat',
		            desc = 'Closes Sanity when you enter combat. This prevents Sanity from getting locked in your way upon entering combat.',
		            get = function()
		            	return Sanity.db.profile.CloseOnCombat
		            end,
		            set = function(v)
		            	Sanity.db.profile.CloseOnCombat = v
		            end
		        },		            	
			      linkcolor = {
			          type = 'toggle',
			          name = 'Link Coloring',
			          desc = 'Use item rarity to color links',
			          get = function()
			          	return Sanity.db.profile.UseColors
			          end,
			          set = function(v)
			          	Sanity.db.profile.UseColors = v
			          end
			      },        	
		        alpha = {
		            type = 'color',
		            name = 'Background Color',
		            desc = "Change Sanity's background",
		            hasAlpha = true,
		            get = function()
		            	return Sanity.db.profile.bg_r,
		            			Sanity.db.profile.bg_g,
		            			Sanity.db.profile.bg_b,
		            			Sanity.db.profile.Alpha
		            end,
		            set = function(r,g,b,a)
		            	Sanity.db.profile.bg_r = r
		            	Sanity.db.profile.bg_g = g
		            	Sanity.db.profile.bg_b = b
		            	Sanity.db.profile.Alpha = a
		            	Sanity:UpdateBackground()
		            end
		        },
		        itemheight = {
		            type = 'range',
		            name = 'Item Height',
		            min = 15,
		            max = 150,
		            step = 1,
		            isPercent = false,
		            desc = "Change the height of each item",
		            get = function()
		            	return Sanity.db.profile.ItemHeight
		            end,
		            set = function(v)
		            	Sanity.db.profile.ItemHeight = v
		            	Sanity:SetupItems()
		            end
		        },        
		        headers = {
		            type = 'toggle',
		            name = 'Show Header',
		            desc = 'Toggle item headers on and off',
		            get = function()
		            	return Sanity.db.profile.UseHeaders
		            end,
		            set = function(v)
		            	Sanity.db.profile.UseHeaders = v
		            	Sanity:Refresh()
		            end
		        },                
		        scale = {
		            type = 'range',
		            name = 'Scale',
		            step = 0.05,
		            min = 0.1,
		            max = 2.0,
		            isPercent = true,
		            desc = "Changes the scale of the Sanity window",
		            get = function()
		            	return Sanity.db.profile.Scale
		            end,
		            set = function(v)
		            				if Sanity.db.profile.Scale ~= v then
			            				Sanity.db.profile.Scale = v
		  	          				SanityFrame:SetScale(v)
													SanityFrame:ClearAllPoints();
													SanityFrame:SetPoint("CENTER", "UIParent", "CENTER");
												end
		            			end
		        },			      
        	}
        },
        usebags = {
        	type = "group",
        	name = "Include Bags",
        	desc = "Bags to include in the display",
        	args = {
			      usebag0 = {
			          type = 'toggle',
			          name = 'Include Backpack',
			          desc = "Scan the backpack and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanBag0
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanBag0 = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      },
			      usebag1 = {
			          type = 'toggle',
			          name = 'Include Bag 1',
			          desc = "Scan Bag #1 and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanBag1
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanBag1 = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      },
			      usebag2 = {
			          type = 'toggle',
			          name = 'Include Bag 2',
			          desc = "Scan Bag #2 and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanBag2
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanBag2 = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      },
			      usebag3 = {
			          type = 'toggle',
			          name = 'Include Bag 3',
			          desc = "Scan Bag #3 and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanBag3
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanBag3 = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      },
			      usebag4 = {
			          type = 'toggle',
			          name = 'Include Bag 4',
			          desc = "Scan Bag #4 and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanBag4
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanBag4 = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      },			      			      			      
			      usekeyring = {
			          type = 'toggle',
			          name = 'Include Keyring',
			          desc = "Scan the key ring and include it in Sanity's display.",
			          get = function()
			          	return Sanity.db.profile.ScanKeyRing
			          end,
			          set = function(v)
			          	Sanity.db.profile.ScanKeyRing = v
			          	SanityItemCache:BAG_UPDATE()
			          	Sanity.displayListDirty = true
			         		Sanity:Refresh()
			          end
			      }                	
        	}
        },
    },
}

Sanity.OnMenuRequest = options
Sanity.name = "Sanity"
Sanity.hasIcon = true


function Sanity:OnClick()	
	Sanity:Toggle()
end

function Sanity:OnTextUpdate()
	self:SetIcon("Interface\\Buttons\\Button-Backpack-Up")
	if (SanityItemCache.db.realm.chars[UnitName("player")].Bags.filled_slots) then
		self:SetText(
			SanityItemCache.db.realm.chars[UnitName("player")].Bags.filled_slots .. " / " .. SanityItemCache.db.realm.chars[UnitName("player")].Bags.bag_slots
		)
	else
		self:SetText("Sanity")
	end
end

function Sanity:OnTooltipUpdate()
		local r,g,b = crayon:GetThresholdColor(1 - (SanityItemCache.db.realm.chars[UnitName("player")].Bags.filled_slots / SanityItemCache.db.realm.chars[UnitName("player")].Bags.bag_slots))
    local cat = tablet:AddCategory(
        'text', "Bags",
        'columns', 1,
        'child_textR', r,
        'child_textG', g,
        'child_textB', b
    )
    
    cat:AddLine(
        'text', SanityItemCache.db.realm.chars[UnitName("player")].Bags.filled_slots .. "/" .. SanityItemCache.db.realm.chars[UnitName("player")].Bags.bag_slots
    )
    
    r,g,b = crayon:GetThresholdColor(1 - (SanityItemCache.db.realm.chars[UnitName("player")].Bank.filled_slots / SanityItemCache.db.realm.chars[UnitName("player")].Bank.bag_slots))
--    Sanity:Print(r .. ", " .. g .. ", " .. b)
    local cat = tablet:AddCategory(
        'text', "Bank",
        'columns', 2,
        'child_textR', r,
        'child_textG', g,
        'child_textB', b
    )
    
    cat:AddLine(
        'text', SanityItemCache.db.realm.chars[UnitName("player")].Bank.filled_slots .. "/" .. SanityItemCache.db.realm.chars[UnitName("player")].Bank.bag_slots
    )
    
    tablet:SetHint("Click to toggle Sanity")
    -- as a rule, if you have an OnClick or OnDoubleClick or OnMouseUp or OnMouseDown, you should set a hint.
end

function Sanity:menu(arg1, arg2, arg3, arg4, arg5, arg6)
	if arg1 == nil then	-- level 1
		dewdrop:AddLine("text", Sanity.menuItem["Name"], 
			"isTitle", true,
			"icon", Sanity.menuItem["Texture"]
		)
		dewdrop:AddLine("text", "Category", "isTitle", true)
		local cats = Sanity:GetExistingCategories()
		dewdrop:AddLine("text", "Existing Categories", "hasArrow", true, "value", "existing_categories", "disabled", #cats == 0)
		dewdrop:AddLine("text", "New Category", 
			"hasArrow", true, 
			"hasEditBox", true, 
			"editBoxText", Sanity.menuItem["TypeSort"], 
			"editBoxFunc", function(item, text) Sanity:SetItemCustomCategory(item, text); dewdrop:Close() end, 
			"editBoxArg1", Sanity.menuItem["Name"],
			"tooltipTitle", "Existing Categories",
			"tooltipText", "A list of custom categories you've already put items into. Select an item from this category to set " .. Sanity.menuItem["Name"] .. "'s category."
		)
		dewdrop:AddLine("text", "Set To Default", 
			"func", function(item) Sanity:SetItemCustomCategory(item, nil); dewdrop:Close() end,
			"arg1", Sanity.menuItem["Name"],
			"tooltipTitle", "Set To Default",
			"tooltipText", "Clears any custom category you've set."
		)
		
		dewdrop:AddLine("text", "Tags", "isTitle", true)
		local cats = Sanity:GetTagList()
		dewdrop:AddLine("text", "Existing Tags", "hasArrow", true, "value", "existing_tags", "disabled", #cats == 0)
		dewdrop:AddLine("text", "New Tag", 
			"hasArrow", true, 
			"hasEditBox", true, 
			"editBoxText", "", 
			"editBoxFunc", function(item, text) Sanity:TagItem(item, text); dewdrop:Close() end, 
			"editBoxArg1", Sanity.menuItem["Name"]
		)
		
	elseif arg1 then
		if arg1 == "existing_categories" then
			local cats = Sanity:GetExistingCategories()
			table.sort(cats)
			for k,v in pairs(cats) do
				dewdrop:AddLine("text", v, "isRadio", true, "checked", Sanity.menuItem["TypeSort"] == v,
					"func", function(item, text) Sanity:SetItemCustomCategory(item, text); dewdrop:Close() end,
					"arg1", Sanity.menuItem["Name"],
					"arg2", v
				)
			end
		elseif arg1 == "existing_tags" then
			local allTags = Sanity:GetTagList()
			table.sort(allTags)
			for k,v in pairs(allTags) do
				local hasTag = Sanity:ItemHasTag(Sanity.menuItem["Name"], v)
				dewdrop:AddLine("text", v, "checked", hasTag,
					"func", function(val, item, text)
						if val then 
							Sanity:TagItem(item, text);
						else
							Sanity:UntagItem(item, text);
						end
					end,
					"arg1", not hasTag,
					"arg2", Sanity.menuItem["Name"],
					"arg3", v
				)
			end
		end
	end
end

function Sanity:searchmenu(arg1, arg2, arg3, arg4, arg5, arg6)
	-- Sanity:Print(tostring(arg1) .. ", " .. tostring(arg2) .. ", " .. tostring(arg3))
	if arg1 == nil then	-- level 1
		local cats = Sanity:GetExistingCategories()
		dewdrop:AddLine("text", "Search", "isTitle", true)
		dewdrop:AddLine("text", "Categories", "hasArrow", true, "value", "existing_categories", "disabled", #cats == 0)
		local cats = Sanity:GetTagList()
		dewdrop:AddLine("text", SANITY_TAGS_LABEL, "hasArrow", true, "value", "existing_tags", "disabled", #cats == 0)
		dewdrop:AddLine("text", "Manage", "isTitle", true)
		dewdrop:AddLine("text", SANITY_TAGS_LABEL, "hasArrow", true, "value", "manage_tags", "disabled", #cats == 0)
		dewdrop:AddLine("text", SANITY_CHARS_LABEL, "hasArrow", true, "value", "manage_chars")
	elseif arg2 == "used_on" then
		dewdrop:AddLine("text", "Untag", "func", function(v, tag) Sanity:UntagItem(v, tag) end, "arg1", arg1, "arg2", arg3)
	elseif arg2 == "manage_chars" then
		dewdrop:AddLine("text", "Delete", "disabled", arg1 == UnitName("player"), "func", function(v)
			StaticPopupDialogs["SANITY_CONFIRM_CHARACTER_DELETE"].text = "Are you sure you want to delete " .. v .. "'s Sanity data?"
			StaticPopupDialogs["SANITY_CONFIRM_CHARACTER_DELETE"].OnAccept = function()
				SanityItemCache:NukeCharacter(v);
				PlaySound("JewelcraftingFinalize");
			end
			StaticPopup_Show("SANITY_CONFIRM_CHARACTER_DELETE")
		end, "arg1", arg1)
	elseif arg1 == "used_on" then
		local items = Sanity:GetItemsForTag(arg2)
		table.sort(items)
		local count = table.getn(items)
		local ct = 0
		for k,v in pairs(items) do
			dewdrop:AddLine("text", v, "hasArrow", true, "value", v)
			ct = ct + 1
			if ct >= 25 and count > 25 then 
				dewdrop:AddLine("text", (count - ct) .. " more...")
				break
			end
		end
	elseif arg1 and arg2 == "manage_tags" then
		local items = Sanity:GetItemsForTag(arg1)
		local i = SANITY_ITEMS_PLURAL
		if #items == 1 then i = SANITY_ITEMS_SINGULAR end
		dewdrop:AddLine("text", SANITY_TAGS_USEDON .. " " .. #items .. " " .. i, "hasArrow", true, "value", "used_on", "disabled", #items == 0)
		dewdrop:AddLine("text", "Delete", "func", function(v, text) Sanity:KillTag(v) end, "arg1", arg1)
	elseif arg1 then
		if arg1 == "existing_categories" then
			local cats = Sanity:GetExistingCategories()
			table.sort(cats)
			for k,v in pairs(cats) do
				dewdrop:AddLine("text", v, "isRadio", true, "checked", Sanity.menuItem["TypeSort"] == v,
					"func", function(text) SanityEditBox:SetText(text); dewdrop:Close() end,
					"arg1", v
				)
			end
		elseif arg1 == "existing_tags" then
			local allTags = Sanity:GetTagList()
			table.sort(allTags)
			for k,v in pairs(allTags) do
				dewdrop:AddLine("text", v, 
					"func", function(text) SanityEditBox:SetText(text); dewdrop:Close() end,
					"arg1", v
				)
			end
		elseif arg1 == "manage_tags" then
			local allTags = Sanity:GetTagList()
			table.sort(allTags)
			for k,v in pairs(allTags) do
				dewdrop:AddLine("text", v, "hasArrow", true, "value", v)
			end
		elseif arg1 == "manage_chars" then
			local chars = {}
			for k,v in pairs(Sanity.ItemCache.db.realm.chars) do
				tinsert(chars, k)
			end
			table.sort(chars)
			for k,v in pairs(chars) do
				dewdrop:AddLine("text", v, "hasArrow", true, "value", v)
			end
		end
	end
end

--[[ Functions ]]--

function Sanity:Register()
	Sanity:RegisterChatCommand({"/sanity", "/san"}, options)
end

function Sanity:ToggleOptions(frame)
	if dewdrop:IsOpen() then
		dewdrop:Close()
	else
		dewdrop:Open(frame, 'children', 
			function()
				dewdrop:FeedAceOptionsTable(options)
			end
		)
	end
end

function Sanity:OnItemEnter(item, anchor)
	if not item or (item and item.invalid and not Sanity.db.profile.UseDangerousItems) then return end
	if item["Link"] == Sanity.lastTooltipLink then GameTooltip:Show(); return end
	Sanity.lastTooltipLink = item["Link"]
	GameTooltip:Hide()
	GameTooltip:SetOwner(anchor, "ANCHOR_LEFT");
	if item["Owner"] == UnitName("player") then
		if item["Bag"] == -1 or item["Bag"] >= 5 then
			if Sanity.BankIsVisible and item["Bag"] == -1 then
				GameTooltip:SetInventoryItem("player", BankButtonIDToInvSlotID(item["Slot"], nil));
			elseif Sanity.BankIsVisible then
				ContainerFrameItemButton_OnEnter()
			else
				GameTooltip:ClearLines();
				GameTooltip:SetHyperlink(item["Link"])
			end
		elseif item["Bag"] == SanityItemCache.INVENTORY_BAG_ID then
			GameTooltip:SetInventoryItem("player", item["Slot"]);
		elseif item["Bag"] ~= SanityItemCache.MAILBOX_BAG_ID then
			local oldID = this:GetID()
			local pOldID = this:GetParent():GetID()
			this:SetID(item["Slot"])
			this:GetParent():SetID(item["Bag"])
			ContainerFrameItemButton_OnEnter();
			this:SetID(oldID)
			this:GetParent():SetID(pOldID)
		end
	elseif item["Bag"] ~= SanityItemCache.MAILBOX_BAG_ID then
		GameTooltip:ClearLines();
		GameTooltip:SetHyperlink(item["Link"])
	end
end

function Sanity:ToggleFilters(val)
	if not val then
		-- Sanity.db.profile.filters_shown = false
		SanityToggleFiltersFrame:SetPoint("TOPRIGHT", SanityFrame, "TOPLEFT", 4, -4)
		SanityFiltersFrame:Hide()
		SanityToggleFiltersFrameTexture:Show()
	else
		-- Sanity.db.profile.filters_shown = true
		SanityToggleFiltersFrameTexture:Hide()
		SanityFiltersFrame:Show()
		SanityToggleFiltersFrame:SetPoint("TOPRIGHT", SanityFiltersFrame, "TOPLEFT", 4, -4)
	end
end

function Sanity:OpenItemOptions(element, item)
	Sanity.menuItem = item
	dewdrop:Open(element, 'children', Sanity.menu, 'point', 
		function(parent) 
			return "TOPLEFT", "TOPRIGHT"
		end
	)
end

-----------------------------------------------------------
-- Updates the interface with our updated item list.
-----------------------------------------------------------
function Sanity:UpdateDisplay(list)
	if not SanityFrame:IsVisible() then return; end
	Sanity:Debug("Doing update")
	if not list then
		Sanity:Debug("Getting display list")
		list = Sanity:GetDisplayList()
	end
	Sanity.viewList = list

	local itemsList = list;
	local numEntries = getn(itemsList);
	
	FauxScrollFrame_Update(SanityFrameScrollFrame, numEntries, Sanity.visibleItems, Sanity.db.profile.ItemHeight);
	
	local offset = FauxScrollFrame_GetOffset(SanityFrameScrollFrame);
	for i = 1, Sanity.visibleItems do
		local button = Sanity.tabletop:GetRow(i)
		
		local index = i + offset;
		if ( index <= numEntries ) then
			if(not button:IsVisible()) then
				button:Show();
			end
			local buttonHeader	= Sanity.tabletop:GetSectionHeader(i)
			
			local item = itemsList[index]
			button.item = nil	
			button:SetAttribute("unit", ATTRIBUTE_NOOP)
			if item then button:Show() else button:Hide() end
			if item["Priority"] == 1 then
				buttonHeader:SetText(itemsList[index]["Header"]);
				buttonHeader:SetWidth(SanityFrame:GetWidth())
				buttonHeader:SetTextColor(Sanity.defaultColor[1], Sanity.defaultColor[2], Sanity.defaultColor[3], 1);
			else
				local buttonName 			= Sanity.tabletop:GetRowField(i, "Name")
				local buttonQty 			= Sanity.tabletop:GetRowField(i, "TypeSortQty")
				local buttonTexture 	= Sanity.tabletop:GetRowField(i, "TypeSort")
				local buttonLocation	= Sanity.tabletop:GetRowField(i, "Location")
				local buttonOwner	 		= Sanity.tabletop:GetRowField(i, "Owner")
				local buttonTags 		  = Sanity.tabletop:GetRowField(i, "Tags")
				local buttonStacks 		  = Sanity.tabletop:GetRowField(i, "Stacks")
			
				button.count = item["Count"]
				button.item = item
				if item["Slot"] then
					button:SetAttribute("unit", "player")
					button:SetAttribute("type2", "item")
					button:SetAttribute("bag", item["Bag"])
					button:SetAttribute("slot", item["Slot"])
					button:SetParent("SanityDummyBagFrame" .. item["Bag"])
					button:SetID(item["Slot"])
				end
				buttonName:SetText(item["Name"]);
				if Sanity.db.profile.UseColors then
					if item["Quality"] then
						local c = Sanity.ItemColors[item["Quality"]+1];
						buttonName:SetTextColor(c[1],c[2],c[3],1);
					else
						buttonName:SetTextColor(1, 0, 0, 1);
					end
				else
					buttonName:SetTextColor(Sanity.defaultColor[1], Sanity.defaultColor[2], Sanity.defaultColor[3], 1);
				end
				buttonQty:SetText(item["Count"]);
				buttonTexture:SetNormalTexture(item["Texture"]);
				buttonOwner:SetText(item["Owner"])
				buttonLocation:SetText(item["Location"])
				buttonTags:SetText(item["Tags"])
				buttonStacks:SetText(item["Stacks"])
			end
		else
			button:Hide();
		end
	end
	local filled_slots = SanityItemCache.db.realm.chars[Sanity.character].Bags.filled_slots or 0
	local bag_slots = SanityItemCache.db.realm.chars[Sanity.character].Bags.bag_slots or 0
	local div_slots = SanityItemCache.db.realm.chars[Sanity.character].Bags.bag_slots or 1
	
	SanityFrameBagSlots:SetText(SANITY_TABS_BAG .. ": " .. filled_slots .. "/" .. bag_slots)
	SanityFrameBagSlots:SetTextColor(tonumber(filled_slots) / tonumber(div_slots), 1 - (tonumber(filled_slots) / tonumber(div_slots)), 0, 1)

	if SanityItemCache.db.realm.chars[Sanity.character].Bank then
		filled_slots = SanityItemCache.db.realm.chars[Sanity.character].Bank.filled_slots or 0
		bag_slots = SanityItemCache.db.realm.chars[Sanity.character].Bank.bag_slots or 0
		div_slots = SanityItemCache.db.realm.chars[Sanity.character].Bank.bag_slots or 1
	else
		filled_slots = 0
		bag_slots = 0
		div_slots = 1
	end
		
	SanityFrameBankSlots:SetText(SANITY_TABS_BANK .. ": " .. filled_slots .. "/" .. bag_slots)
	SanityFrameBankSlots:SetTextColor(
		tonumber(filled_slots) / tonumber(div_slots), 
		1 - (tonumber(filled_slots) / tonumber(div_slots)), 
		0, 
		1)
end

function Sanity:ShowLastItemTagMenu()
	dewdrop:Close()
	if not Sanity.lastFocusName or Sanity.lastFocus ~= GetMouseFocus() then
		return
	end
	Sanity.menuItem = {
		["Name"] = Sanity.lastFocusName,
		["Texture"] = Sanity.lastFocusTexture,
		["TypeSort"] = Sanity:GetCategoryFor(Sanity.lastFocusName) or ""
	}
	dewdrop:Open(Sanity.lastFocus, 'children', Sanity.menu, 'point', 
		function(parent) 
			local x,y = GetCursorPosition()
			x = x / UIParent:GetEffectiveScale()
			y = y / UIParent:GetEffectiveScale()
			local xr = "LEFT"
			local yr = "TOP"
			local px = "RIGHT"
			local py = "TOP"
			if x > (UIParent:GetWidth() / 2) then
				xr = "RIGHT"
				px = "LEFT"
			end
			if y < (UIParent:GetHeight() / 2) then
				yr = "BOTTOM"
				py = "BOTTOM"
			end
			return (yr .. xr), (py .. px)
		end
	)
end

StaticPopupDialogs["SANITY_CONFIRM_CHARACTER_DELETE"] = {
	text = "Are you sure you want to delete <char>?",
	button1 = TEXT(YES),
	button2 = TEXT(NO),
	timeout = 0,
	showAlert = 1,
	hideOnEscape = 1,
};

StaticPopupDialogs["SANITY_CONFIRM_DATABASE_WIPE"] = {
	text = "Are you sure you want to reset your Sanity settings?",
	button1 = TEXT(YES),
	button2 = TEXT(NO),
	timeout = 0,
	showAlert = 1,
	hideOnEscape = 1,
};

StaticPopupDialogs["SANITY_CONFIRM_DATABASE_WIPE"].text = "Are you sure you want to delete all of Sanity's saved settings and data?"
StaticPopupDialogs["SANITY_CONFIRM_DATABASE_WIPE"].OnAccept = function()
	Sanity:ResetDB()
	SanityItemCache:ResetDB();
	PlaySound("JewelcraftingFinalize");
	ReloadUI()
end
